# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit

from aiida_fleur.workflows.ssdisp import FleurSSDispWorkChain

structure = load_node(STRUCTURE_PK)
fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)

wf_para = Dict(
    dict={
        'beta': {
            'all': 1.57079
        },
        'prop_dir': [0.125, 0.125, 0.0],
        'q_vectors': [[0.0, 0.0, 0.0], [0.125, 0.125, 0.0], [0.250, 0.250, 0.0], [0.375, 0.375, 0.0],
                      [0.500, 0.500, 0.0]],
        'ref_qss': [0.0, 0.0, 0.0],
        'inpxml_changes': [],
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        }
    })

options = Dict(
    dict={
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 24
        },
        'queue_name': 'devel',
        'custom_scheduler_commands': '',
        'max_wallclock_seconds': 60 * 60
    })

parameters = Dict(
    dict={
        'atom': {
            'element': 'Pt',
            'lmax': 8
        },
        'atom2': {
            'element': 'Fe',
            'lmax': 8,
        },
        'comp': {
            'kmax': 3.8,
        },
        'kpt': {
            'div1': 20,
            'div2': 24,
            'div3': 1
        }
    })

wf_para_scf = {'fleur_runmax': 2, 'itmax_per_run': 120, 'density_converged': 0.2, 'mode': 'density'}

wf_para_scf = Dict(dict=wf_para_scf)

options_scf = Dict(
    dict={
        'resources': {
            'num_machines': 2,
            'num_mpiprocs_per_machine': 24
        },
        'queue_name': 'devel',
        'custom_scheduler_commands': '',
        'max_wallclock_seconds': 60 * 60
    })

inputs = {
    'scf': {
        'wf_parameters': wf_para_scf,
        'structure': structure,
        'calc_parameters': parameters,
        'options': options_scf,
        'inpgen': inpgen_code,
        'fleur': fleur_code
    },
    'wf_parameters': wf_para,
    'fleur': fleur_code,
    'options': options
}

res = submit(FleurSSDispWorkChain, **inputs)



# -*- coding: utf-8 -*-
{
    'conv_mode': 'density',
    'distance_charge': 0.1406279038,
    'distance_charge_all': [
        61.1110641131,
        43.7556515683,
        ...
    ],
    'distance_charge_units': 'me/bohr^3',
    'errors': [],
    'force_diff_last': 'can not be determined',
    'force_largest': 0.0,
    'info': [],
    'iterations_total': 23,
    'last_calc_uuid': 'b20b5b94-5d80-41a8-82bf-b4d8eee9bddc',
    'loop_count': 1,
    'material': 'FePt2',
    'total_energy': -38166.176928494,
    'total_energy_all': [
        -38166.542950054,
        -38166.345602746,
        ...
    ],
    'total_energy_units': 'Htr',
    'total_wall_time': 245,
    'total_wall_time_units': 's',
    'warnings': [],
    'workflow_name': 'FleurScfWorkChain',
    'workflow_version': '0.4.0'
}



# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit

from aiida_fleur.workflows.relax import FleurRelaxWorkChain

fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)


wf_relax = {'film_distance_relaxation': False,
            'force_criterion': 0.049,
            'relax_iter': 5
            }

wf_relax_scf = {'fleur_runmax': 5,
                'itmax_per_run': 50,
                'alpha_mix': 0.015,
                'relax_iter': 25,
                'force_converged': 0.001,
                'force_dict': {'qfix': 2,
                               'forcealpha': 0.75,
                               'forcemix': 'straight'},
                'inpxml_changes': []
                }

wf_relax = Dict(dict=wf_relax)
wf_relax_scf = Dict(dict=wf_relax_scf)

calc_relax = {'comp': {'kmax': 4.0,
                       },
              'kpt': {'div1': 24,
                      'div2': 20,
                      'div3': 1
                      },
              'atom': {'element': 'Pt',
                       'rmt': 2.2,
                       'lmax': 10,
                       'lnonsph': 6,
                       'econfig': '[Kr] 5s2 4d10 4f14 5p6| 5d9 6s1',
                       },
              'atom2': {'element': 'Fe',
                        'rmt': 2.1,
                        'lmax': 10,
                        'lnonsph': 6,
                        'econfig': '[Ne] 3s2 3p6| 3d6 4s2',
                        },
              }

calc_relax = Dict(dict=calc_relax)

options_relax = {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 4, 'num_cores_per_mpiproc': 6},
                 'queue_name': 'devel',
                 'custom_scheduler_commands': '',
                 'max_wallclock_seconds':  1*60*60}

inputs = {
    'scf': {
        'wf_parameters': wf_relax_scf,
        'calc_parameters': calc_relax,
        'options': options_relax,
        'inpgen': inpgen_code,
        'fleur': fleur_code
    },
    'wf_parameters': wf_relax
}

res = submit(FleurRelaxWorkChain, **inputs)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Here default values are collected.
Meaning is to centralize the defaults throughout the plugin more, to allow in the
end that the user might be able to specify them in a file he provides
'''
default_options = {
    'resources': {
        'num_machines': 1,
        'num_mpiprocs_per_machine': 1
    },
    'max_wallclock_seconds': 6 * 60 * 60,
    'queue_name': '',
    #'custom_scheduler_commands': '',
    #'import_sys_environment': False,
    #'environment_variables': {}
}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains useful function to extract data from nodes,
specific to aiida-fleur. Any useful code snipets for handling large number of
nodes, or data-mine go in here
"""

#import time
from aiida.plugins import DataFactory
from aiida.orm.querybuilder import QueryBuilder as QB


def extract_structure_info(keys, structures=None):
    """
    A method that collects a bunch of information (specified in keys) from
    structures (default whole db, or provided node list) in the database and
    returns that information as a dict, which could be used for further evalation
    #keys = ['uuid', 'formula', 'pk', 'symmetry', 'pbc', 'volume', 'total_energy',
    'child_nodes', 'natoms', 'group', extras', 'label', 'description', 'cif_file',
    'cif_number', 'cif_uuid', 'cif_ref', 'calcfunctions', 'band', 'dos', 'eos',
    'init_cls', 'corehole', primitive]

    """
    StructureData = DataFactory('structure')
    structure_list = []

    from aiida_fleur.tools.StructureData_util import get_spacegroup, is_structure
    from aiida_fleur.tools.StructureData_util import is_primitive

    if not structures:
        StructureData = DataFactory('structure')
        #t = time.time()
        qb = QB()
        qb.append(StructureData)
        structures = qb.all()
        #elapsed = time.time() - t
        # print "Total number of structures: {} (retrieved in {} s.)".format(len(structures), elapsed)
        #t = time.time()

    # for structure in structures:
    #    structure_dict = {}
    #    struc = structure[0]
    #    for key in keys:
    #        structure_dict[key] = get_methods(key)(struc)

    # get information
    for structure in structures:
        structure_dict = {}

        if isinstance(structure, list):
            struc = structure[0]
        else:
            struc = is_structure(structure)

        if 'formula' in keys:
            structure_dict['formula'] = struc.get_formula()
        if 'pk' in keys:
            structure_dict['pk'] = struc.pk
        if 'uuid' in keys:
            structure_dict['uuid'] = str(struc.uuid)
        if 'natoms' in keys:
            structure_dict['natoms'] = len(struc.sites)
        if 'cell' in keys:
            structure_dict['cell'] = str(struc.cell)
        if 'pbc' in keys:
            structure_dict['pbc'] = str(struc.pbc)
        if 'label' in keys:
            structure_dict['label'] = struc.label
        if 'description' in keys:
            structure_dict['description'] = struc.description
        if 'extras' in keys:
            extras = struc.extras
            structure_dict['extras'] = str(extras)
        if 'symmetry' in keys:
            symmetry = get_spacegroup(struc)
            structure_dict['symmetry'] = str(symmetry)
        if 'volume' in keys:
            volume = struc.get_cell_volume()
            structure_dict['volume'] = volume
        if 'child_nodes' in keys:
            child_nodes = len(struc.get_outgoing().all())
            structure_dict['child_nodes'] = child_nodes
        if 'primitive' in keys:
            prim = is_primitive(struc)
            structure_dict['primitive'] = prim

        if 'cif_file' in keys:
            cif_file = get_cif_file(struc)
            structure_dict['cif_file'] = cif_file
        '''
        if 'cif_number' in keys:
            cif_number = get_cif_number(struc)
            structure_dict['cif_number'] = cif_number
        if 'cif_uuid' in keys:
            cif_uuid = get_cif_uuid(struc)
            structure_dict['cif_uuid'] = cif_uuid
        if 'cif_ref' in keys:
            cif_ref = get_cif_ref(struc)
            structure_dict['cif_ref'] = cif_ref
        if 'total_energy' in keys:
            total_energy = get_total_energy(struc)
            structure_dict['total_energy'] = total_energy
        '''
        if 'group' in keys:
            group = group_member(struc)
            structure_dict['group'] = group
        if 'scf' in keys:
            scf = input_of_workcal('fleur_scf_wc', struc)
            structure_dict['scf'] = scf
        if 'band' in keys:
            band = input_of_workcal('fleur_band_wc', struc)
            structure_dict['band'] = band
        if 'dos' in keys:
            dos = input_of_workcal('fleur_dos_wc', struc)
            structure_dict['dos'] = dos
        if 'eos' in keys:
            eos = input_of_workcal('fleur_eos_wc', struc)
            structure_dict['eos'] = eos
        if 'init_cls' in keys:
            init_cls = input_of_workcal('FleurInitialCLSWorkChain', struc)
            structure_dict['init_cls'] = init_cls
        if 'corehole' in keys:
            corehole = input_of_workcal('', struc)
            structure_dict['corehole'] = corehole
        if 'calcfunctions' in keys:
            calcfunctions_uuid, calcfunctions_name = input_of_calcfunctions(struc)
            structure_dict['calcfunctions'] = [calcfunctions_uuid, calcfunctions_name]

        structure_list.append(structure_dict)

    #elapsed = time.time() - t
    # print "(needed {} s.!!!)".format(elapsed)

    return structure_list


def group_member(node):
    """
    Find to what groups a node belongs to.

    Comment: currently very greedy!
    """
    from aiida.orm import Group
    member_in = []
    # get all groups in db
    # for each group check if node is member of group
    # append group name to member_in

    res = Group.objects.all()
    for group in res:
        name = group.label
        for gnode in group.nodes:  # TODO: easier/better way?
            if gnode.uuid == node.uuid:
                member_in.append(name)
    return member_in


def input_of_workcal(name, node):
    """
    checks if a given node was input into a certain WorkChain
    and returns a list of workcalculation uuids of workcalculations with the given name
    """
    from aiida.orm import WorkChainNode
    process_uuids = []
    for out in node.get_outgoing().all_nodes():
        if isinstance(out, WorkChainNode):
            label = out.get_attribute('process_label')
            if label == name:
                process_uuids.append(out.uuid)
    return process_uuids


def input_of_calcfunctions(node, name=''):
    """
    checks if a given node was input into a certain calcfunction
    and returns a list of calcfunction uuids of calcfunction with the given name
    """
    from aiida.orm import CalcFunctionNode
    process_uuids = []
    process_names = []
    outputs = node.get_outgoing().all_nodes()
    for out in outputs:
        if isinstance(out, CalcFunctionNode):
            try:  # TODO: is there a better way
                label = out.get_attribute('process_label')
            except AttributeError:
                label = None
                continue
            if label == name:
                process_uuids.append(out.uuid)
                process_names.append(label)
            if not name:
                process_uuids.append(out.uuid)
                process_names.append(label)
    return process_uuids, process_names


def get_cif_file(node):
    """
    Finds out if (a structure) given as input was created from a cif file
    currently with the method wf_struc_from_cif

    params: node: structureData node
    returns [cif_filename, cif_uuid]
    """
    from aiida.orm import CifData, CalcFunctionNode
    inputs = node.get_incoming().all_nodes()
    name = 'wf_struc_from_cif'  # TODO: Bad solution, not general for me currently enough
    cif_uuid, cif_filename = '', ''
    for inp in inputs:
        if isinstance(inp, CalcFunctionNode):
            try:  # TODO: is there a better way
                label = inp.get_attribute('process_label')
            except AttributeError:
                label = None
                continue
            if label == name:
                inp_wc = inp.get_incoming().all_nodes()
                for iwc in inp_wc:
                    if isinstance(iwc, CifData):
                        cif_uuid = iwc.uuid
                        cif_filename = iwc.filename
                        break

    return [cif_filename, cif_uuid]


# wf_struc_from_cif process_label
# cif.filename
# cif.uuid
# cif.folder.abspath
# wc.get_incoming().all()



# -*- coding: utf-8 -*-
'beta': {'all': 1.57079},                   # see the description below
'q_vectors': {'label': [0.0, 0.0, 0.0],     # sets q_points to calculate
              'label2': [0.125, 0.0, 0.0]
              }
'suppress_symmetries': False                # True if use no symmetries



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Contains helper functions to create core-holes in
Fleur input files from AiiDA data nodes.
'''


# TODO maybe merge these methods into fleurinp or structure util? or create a parameterData utils
#355
def create_corehole_para(structure, kind, econfig, species_name='corehole', parameterdata=None):
    """
    This methods sets of electron configurations for a kind
    or position given, make sure to break the symmetry for this position/kind
    beforehand, otherwise you will create several coreholes.

    :param structure: StructureData
    :param kind: a string with the kind_name (TODO: alternative the kind object)
    :param econfig: string, e.g. econfig = "[Kr] 5s2 4d10 4f13 | 5p6 5d5 6s2" to set, i.e. the corehole

    :return: a Dict node
    """
    # TODO: Since fleur MaXR5 there is a default econfig file and the order behavior
    # has changed. now to atom lists only change the default if they have an id.
    from aiida.common.constants import elements as PeriodicTableElements
    from aiida import orm
    _atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}
    #from aiida_fleur.tools.merge_parameter import merge_parameter

    kindo = structure.get_kind(kind)
    symbol = kindo.symbol
    head = kindo.name.rstrip('01223456789')
    charge = _atomic_numbers[kindo.symbol]
    a_id = float(f'{charge}.{kindo.name[len(head):]}')

    # get kind symbol, get kind name,
    #&atom element="W" jri=921 lmax=8 rmt=2.52 dx=0.014 lo="5p" econfig="[Kr] 5s2 4d10 4f13 | 5p6 5d4 6s2" /
    #count = 0
    if parameterdata:
        new_parameterd = parameterdata.get_dict()  # dict()otherwise parameterdata is changed
        for key, val in new_parameterd.items():
            if 'atom' in key:
                if val.get('element', None) == symbol:
                    # remember atomic id is atomic number.some int
                    if (a_id and float(a_id) == float(val.get('id', -1))):
                        val.update({'econfig': econfig})
                        break
                    if not a_id:
                        val.update({'econfig': econfig})
                    else:
                        pass
    else:
        if a_id:
            if species_name:
                new_parameterd = {'atom': {'element': symbol, 'econfig': econfig, 'id': a_id, 'name': species_name}}
            else:
                new_parameterd = {'atom': {'element': symbol, 'econfig': econfig, 'id': a_id}}
        else:
            new_parameterd = {'atom': {'element': symbol, 'econfig': econfig}}

    new_parameter = orm.Dict(dict=new_parameterd)
    #if parameterdata:
    #    new_parameter = merge_parameter(parameterdata, new_parameter)
    return new_parameter  #structure


'''
# Move to fleurinpmod? fleurinp->self
# This method is fully implemented yet since it turned out to better go over inpgen
def create_corehole_fleurinp(fleurinp, species, stateocc, pos=None, coreconfig='same', valenceconfig='same'):
    """
    Removes an electron from the core and adds it to the valence band of the kind
    given econfig as in inp.xml::

        [Kr] (5s1/2) (4d3/2) (4d5/2) (4f5/2) (4f7/2)

    if position(pos) is given the electronConfig for the specified position will be set.
    (or todo? econfig, either::

            [Kr] 5s2 4d10 4f13 | 5p6 5d4 6s2

    or::

        [Kr] 2(5s1/2) 4(4d3/2) 6(4d5/2) 6(4f5/2) 8(4f7/2) |2(5p1/2) 4(5p3/2) 2(6s1/2) 2(5d3/2) 2(5d5/2))

    occ tags already there will be untouched, unless the state is the same as given.

    :params fleurinp: an unstored! changes are done on this fleurinp fleurinpdata object # TODO alternatively stored?
    :params species: string with species name
    :params stateocc: dict state tuples (spinup, spindown), exp: {'(5d3/2)' : (2.5, 0.0), '(4f7/2)' : (3.5 , 4.0)}
    :params pos: list of tuples of 3, pos=[(0.0, 0.0, 0.0), ...]
    :params coreconfig: string, e.g: "[Kr] (5s1/2) (4d3/2) (4d5/2) (4f5/2) (4f7/2)", default='same' (same as current in inp.xml)
    :params valenceconfig string, e.g.: (5p1/2) (5p3/2) (6s1/2) (5d3/2) (5d5/2)

    :return: the changes fleurinpData object
    """

    #     <electronConfig>
    #        <coreConfig>[Kr] (5s1/2) (4d3/2) (4d5/2) (4f5/2) (4f7/2)</coreConfig>
    #        <valenceConfig>(5p1/2) (5p3/2) (6s1/2) (5d3/2) (5d5/2)</valenceConfig>
    #        <stateOccupation state="(5d3/2)" spinUp="2.00000000" spinDown=".00000000"/>
    #        <stateOccupation state="(5d5/2)" spinUp="2.00000000" spinDown=".00000000"/>
    #     </electronConfig>

    from aiida_fleur.tools.xml_util import eval_xpath2, get_xml_attribute
    #from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
    # or from fleurinp?

    FleurinpData = DataFactory('fleur.fleurinp')
    ########### all xpath maintain ########### ? needed?
    electronConfig_xpath = '/fleurInput/atomSpecies/species/electronConfig'
    species_xpath = '/fleurInput/atomSpecies/species'
    #####################

    # test input, if something wrong return/ throw error
    # get electronConfig tag from fleurinp
    # create new tag from old tag and method input
    # return new fleurinp data
    # Best use fleurinp functions
    # Do it with xml or rather

    new_core = False
    new_valence = False

    # test input.
    if not isinstance(fleurinp, FleurinpData):
        print('No fleurinp Data given to "create_valence_corehole"')
        return None  # TODO throw error?

    if coreconfig != 'same':
        new_core = True

    if valenceconfig != 'same':
        new_valence = True
    if pos is None:
        pos = []
    species_tags = fleurinp.get_tag(species_xpath)
    for speci in species_tags:
        if get_xml_attribute(speci, 'name') == species:
            # change
            econfig = eval_xpath2(speci, 'electronConfig')[0]
            coreconfig = eval_xpath2(econfig, 'coreConfig')
            valenceconfig = eval_xpath2(econfig, 'valenceConfig')
            occupations = eval_xpath2(econfig, 'stateOccupation')

            for key, val in six.iteritems(stateocc):
                added = False
                for occ in occupations:
                    name = get_xml_attribute(occ, 'state')
                    if name == key:
                        added = True
                        # override and break (can occur only once)
                        occ.set('spinUp', str(val[0]))
                        occ.set('spinDown', str(val[0]))
                        break
                if not added:
                    pass

    #st_inpchanges(change_dict)

    #alternative
    change_dict = {'coreConfig': '', 'valenceConfig': ''}
    change_dict2 = {'state': '', 'spinUp': '', 'spinDown': ''}
    change_dict3 = {'valenceElectrons': ''}

    return fleurinp


def write_change(xmltree, changelist_xpath):
    """
    applies the changes from the changelist to the xml tree
    """
    xmltree_new = xmltree
    for element in changelist_xpath:
        xpath = element[0]
        value = element[1]
    return xmltree_new
'''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Module with CLI commands for calcjob types of aiida-fleur.
'''
import click
from .launch import launch_inpgen
from .launch import launch_fleur
from .launch import launch_scf
from .launch import launch_banddos
from .launch import launch_relax
from .launch import launch_eos
from .launch import launch_corehole
from .launch import launch_init_cls
from .launch import launch_mae
from .launch import launch_create_magnetic
from .launch import launch_dmi
from .launch import launch_ssdisp


@click.group('launch')
def cmd_launch():
    """Commands to launch workflows and calcjobs of aiida-fleur."""


# we do it like this and not in short from to avoid cyclic imports
# and get the full bash completion working
cmd_launch.add_command(launch_inpgen)
cmd_launch.add_command(launch_fleur)
cmd_launch.add_command(launch_scf)
cmd_launch.add_command(launch_banddos)
cmd_launch.add_command(launch_relax)
cmd_launch.add_command(launch_eos)
cmd_launch.add_command(launch_corehole)
cmd_launch.add_command(launch_init_cls)
cmd_launch.add_command(launch_mae)
cmd_launch.add_command(launch_create_magnetic)
cmd_launch.add_command(launch_dmi)
cmd_launch.add_command(launch_ssdisp)



# -*- coding: utf-8 -*-
'element': '',                          # detemines for which element to calculate
                                        # the crystal field coefficients
'rare_earth_analogue': True,            # True if analogue calculation should be used
'analogue_element': 'Y',                # Which element to use for the analogue structure
'replace_all': True,                    # Whether to replace all atoms for the analogue in one structure
'soc_off': True,                        # if True the socscale is set to 0.0 for the 4f site
'convert_to_stevens': True              # if True the coefficients are converted to the stevens convention
                                        # A_lm<r^l>



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
This module contains click option types specific to aiida-fleur
'''
import click
from aiida.cmdline.params import types
from aiida.cmdline.utils import echo
from aiida.common.exceptions import NotExistent
from aiida.plugins import DataFactory
from aiida.cmdline.utils.decorators import with_dbenv


class StructureNodeOrFileParamType(click.ParamType):
    """
    The ParamType for identifying a structure by node or to extract it from a given file

    Pro: It is convenient
    Con: If users only use other formats to launch their workflows it will create many
    more structures in the database.
    """

    name = 'StructureFile'

    @with_dbenv()
    def convert(self, value, param, ctx):
        is_path = False
        # Alternative one could check if int or uuid
        # aiida allows also for shorten uuids
        from aiida.orm import StructureData, QueryBuilder

        try:
            structure = types.DataParamType(sub_classes=('aiida.data:structure',)).convert(value, param, ctx)
        except (NotExistent, click.exceptions.BadParameter) as er:
            echo.echo(f'Tried to load node, could not fine one for {value}. '
                      'I will further check if it is a filepath.')
            is_path = True

        if is_path:
            # If it is a path to a file try to convert the structure
            pathtype = click.Path(exists=True, dir_okay=False, resolve_path=True)
            filename = pathtype.convert(value, param, ctx)
            try:
                import ase.io
            except ImportError:
                echo.echo_critical('You have not installed the package ase. \nYou can install it with: pip install ase')

            try:
                asecell = ase.io.read(filename)
                structure = StructureData(ase=asecell)
            except ValueError as err:
                echo.echo_critical(str(err))
            # do not store structure, since this option is for calculation and workflow
            # input, which will store the structure anyway.

        # do not store again if structure is already there.
        duplicate = QueryBuilder().append(StructureData, filters={'extras._aiida_hash': structure._get_hash()}).first()  # pylint: disable=protected-access

        if duplicate:
            return duplicate[0]
        return structure



# -*- coding: utf-8 -*-
'iterations_fixed': 30,                         #Number of iterations to run with fixed density matrices
'ldau_dict': {'all-Nd': {'l': 3,                #Specifications of the LDA+U parameters to add
                         'U': 6.7,              #Note that the input has to be without LDA+U
                         'J': 0.7,              #for this wokchain to wotk consistently
                         'l_amf': False}},
'use_orbital_occupation': False,                #If True the obtained configurations are used for the
                                                #atomic orbitals and not the spherical harmonics
'fixed_occupations': {'all-Nd': {3: (4,0)}},    #Specifies the occupations for each LDA+U orbital
                                                #for each spin to generate all possible configurations from
'fixed_configurations': None,                   #Alternative way to specify density matrix configurations
                                                #specifies the explicit configurations to use



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module you find a method (read_cif_folder) to read in all .cif files
from a folder and store the structures in the database.
"""

# TODO: links to the structures created in the db from the cif files have to be
# set. Might make a difference for
# structure visualization, because cif file has more information
# also keep connection to ICSD id number

import os

from aiida.plugins import DataFactory
from aiida.engine import calcfunction as cf


def read_cif_folder(path=os.getcwd(),
                    recursive=True,
                    store=False,
                    log=False,
                    comments='',
                    extras='',
                    logfile_name='read_cif_folder_logfile'):
    """
    Method to read in cif files from a folder and its subfolders.
    It can convert them into AiiDA structures and store them.

    defaults input parameter values are:
    path=".", recursive=True, store=False, log=False, comments='', extras=''

    :params: path: Path to the dictionary with the files (default, where this method is called)
    :params: recursive: bool, If True: looks aso in subfolders, if False: just given dir
    :params: store: bool, if True: stores structures in database
    :params: log: bool, if True, writes a logfile with information (pks, and co)
    :params: comments: string: comment to add to the structures
    :params: extras: dir/string/arb: extras added to the structures stored in the db

    """
    # TODO check for duplicates in the database, so that reruning the functions
    # won't import anything else in the database
    cifdata = DataFactory('cif')
    ############ parameters for the user to set ########

    parent_cif_folder = path  # folder path
    store_db = store  # True # store stuff in database?
    write_log = log  # write a logfiles on what was saved
    comment = comments  # comments and extras to add to the structure nodes.
    extra = extras  # helpfull for finding them again in the db
    rek = recursive  # search also in subfolders or only given folder

    #####################
    filenames = []
    filepaths = []
    infofilestring = ('Structure Formula, Structuredata pk, Structure Data uuid, cif-file-path, comment, extras \n')

    #1. get all the files
    if rek:
        for root, dirs, files in os.walk(parent_cif_folder):
            for file1 in files:
                if file1.endswith('.cif'):
                    filenames.append(file1)
                    filepath = os.path.join(root, file1)
                    filepaths.append(filepath)
    else:
        dir_list = os.listdir(parent_cif_folder)
        for filename in dir_list:
            if filename.endswith('.cif'):
                filenames.append(filename)
                filepath = os.path.join(parent_cif_folder, filename)
                filepaths.append(filepath)

    nfiles = len(filenames)
    print(f'{nfiles} cif-files found in folder "{parent_cif_folder}" ')

    structuredatas = []

    #2. read all the files and store stuff.
    saved_count = 0
    saved_count_cif = 0
    filenames2 = []
    structuredatas2 = []
    for i in range(nfiles):
        try:
            new_cif = cifdata.get_or_create(filepaths[i], store_cif=True)
        except (ValueError, AttributeError, ImportError) as emessage:
            print(f'invalid cif file: {filepaths[i]}, the error message was {emessage} ')
            continue
        #print new_cif
        if new_cif[1]:
            saved_count_cif = saved_count_cif + 1
        # do we want to save the structures again, or do we also continue
        #else:
        #    continue
        #asecell = new_cif[0].get_ase()
        #structuredatas.append(DataFactory('structure'))
        filenames2.append(filenames[i])
        #struc = structuredatas[-1](ase=asecell)
        #formula = struc.get_formula()
        if store_db:
            struc = wf_struc_from_cif(new_cif[0])
            formula = struc.get_formula()
            #new_cif.store()
            #struc.store()
            saved_count = saved_count + 1

            # add comment or extras, only possible after storing
            if comment:
                user = struc.user  # we are the creator
                struc.add_comment(comment, user)
            if extra:
                if isinstance(extra, dict):
                    struc.set_extra_many(extra)
                else:
                    struc.set_extra('specification', extra)
            struc.set_extra('formula', formula)
            structuredatas2.append(struc)
        else:
            struc = struc_from_cif(new_cif[0])
            structuredatas2.append(struc)
            formula = struc.get_formula()
        if write_log:
            # This file is a logfile/info file created by 'read_cif_folder'
            # Structure Formula, structuredata pk, Structure Data uuid,
            #'cif-file-path', comment, extras
            # TODO? if not stored write not stored
            if store_db:
                infofilestring = infofilestring + '{} {} {} {} {} {} \n'.format(
                    formula, struc.pk, struc.uuid, filepaths[i], struc.get_comments(), struc.extras)
            else:
                infofilestring = (infofilestring + '{} notstored notstored {}'
                                  'notstored notstored \n'
                                  ''.format(formula, filepaths[i]))

    # write a logfile
    if write_log:
        file1 = os.open(logfile_name, os.O_RDWR | os.O_CREAT)
        os.write(file1, bytes(infofilestring, 'UTF8'))
        os.close(file1)
    print(f'{saved_count_cif} cif-files and {saved_count} structures were saved in the database')

    return structuredatas2, filenames2


@cf
def wf_struc_from_cif(cif):
    return struc_from_cif(cif)


def struc_from_cif(cif):
    asecell = cif.get_ase()
    struc = DataFactory('structure')(ase=asecell)
    return struc


# TODO add this to command line, or better move to aiida-jutools
# ggf add what Roman has done there.
'''
if __name__ == '__main__':
    import argparse
    import json
    #  maybe change names?
    parser = argparse.ArgumentParser(description="Read '.cif' files from the current"
                                     ' folder and store in AiiDA database. If no'
                                     ' arguements are given, read_cif_folder is'
                                     ' using default arguments')

    parser.add_argument('-p',
                        metavar='path',
                        type=str,
                        required=False,
                        action='store',
                        help='specify path as string, if not current folder')

    parser.add_argument('-r',
                        required=False,
                        action='store_true',
                        help='if given, search also in subfolders for .cif files. (os.walk subfolders')

    parser.add_argument('-s', required=False, action='store_true', help='if given, store all structures in database.')

    parser.add_argument('-l', required=False, action='store_true', help='if given, write a logfile')

    parser.add_argument('-c',
                        metavar='comments',
                        type=str,
                        required=False,
                        default='',
                        action='store',
                        help='string, add a comment to the node(s) if stored. '
                        'exp: -c crystal data for my project')

    parser.add_argument('-e',
                        metavar='extras',
                        required=False,
                        type=json.loads,
                        action='store',
                        help='string, or dictionary to add to the node(s) extras'
                        ' if stored. exp: -e {"project" : "myproject", "type" : "simple metal"}')

    args = parser.parse_args()
    if args.p:
        read_cif_folder(path=args.p, recursive=args.r, store=args.s, log=args.l, comments=args.c, extras=args.e)
    else:
        read_cif_folder(recursive=args.r, store=args.s, log=args.l, comments=args.c, extras=args.e)
'''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
# pylint: disable=cyclic-import
"""Command line utilities to create and inspect `Dict` nodes with FLAPW parameters."""
import click

from aiida.cmdline.params import options as aiida_options
from aiida.cmdline.utils import decorators, echo
from ..util import options as fleur_options
from . import cmd_data


@cmd_data.group('parameter')
def cmd_parameter():
    """Commands to create and inspect `Dict` nodes containing FLAPW parameters ('calc_parameters')."""


@cmd_parameter.command('import')
@click.argument('filename', type=click.Path(exists=True))
@click.option('--fleurinp/--no-fleurinp',
              default=False,
              show_default=True,
              help='Store also the fleurinp and the extractor calcfunction in the db.')
@click.option('--show/--no-show', default=True, show_default=True, help='Print the contents from the extracted dict.')
@aiida_options.DRY_RUN()
@decorators.with_dbenv()
def cmd_param_import(filename, dry_run, fleurinp, show):
    """
    Extract FLAPW parameters from a Fleur input file and store as Dict in the db.

    FILENAME is the name/path of the inp.xml file to use.
    """
    from aiida_fleur.data.fleurinp import FleurinpData

    if not filename.endswith('.xml'):
        echo.echo_critical('Error: Currently, we can only extract information from an inp.xml file.')
    fleurinpd = FleurinpData(files=[filename])
    if not fleurinp or dry_run:
        parameters = fleurinpd.get_parameterdata_ncf()
    else:
        parameters = fleurinpd.get_parameterdata()

    if dry_run:
        echo.echo_success('parsed FLAPW parameters')
    else:
        parameters.store()
        echo.echo_success(f'parsed and stored FLAPW parameters<{parameters.pk}>  <{parameters.uuid}>')

    if show:
        echo.echo_dictionary(parameters.get_dict())


# further ideas:
# query for certain FLAPW parameter nodes.
# Example show me all for Si
# query for options nodes
# command to split and merge parameters nodes together based on elements.
@cmd_data.group('options')
def cmd_options():
    """Commands to create and inspect `Dict` nodes containing options."""


@cmd_options.command('create')
@fleur_options.MAX_NUM_MACHINES()
@fleur_options.NUM_MPIPROCS_PER_MACHINE()
@fleur_options.MAX_WALLCLOCK_SECONDS()
@fleur_options.QUEUE_NAME()
@aiida_options.DRY_RUN()
@click.option('--show/--no-show', default=True, show_default=True, help='Print the contents from the options dict.')
@decorators.with_dbenv()
def cmd_option_create(max_num_machines, num_mpiprocs_per_machine, queue, max_wallclock_seconds, dry_run, show):
    """
    Command to create options dict nodes
    """
    from aiida_fleur.common.node_generators import generate_wf_option_node

    optiondict = {}
    optiondict = {
        'resources': {
            'num_machines': max_num_machines,
            'num_mpiprocs_per_machine': num_mpiprocs_per_machine
        },
        'max_wallclock_seconds': max_wallclock_seconds,
        'queue_name': queue
    }

    options = generate_wf_option_node(**optiondict)

    if dry_run:
        echo.echo_success('Created option node.')
    else:
        options.store()
        echo.echo_success(f'Created and stored Options node <{options.pk}>  <{options.uuid}>')

    if show:
        echo.echo_dictionary(options.get_dict())



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Module with CLI commands to launch for calcjob and workflows of aiida-fleur.
'''
# TODO: these launch commands should be put in separate files, if this one becomes to large..

import click
from ..util import options, utils, defaults
from aiida_fleur.tools.dict_util import clean_nones
from aiida.orm import Code, load_node, Dict
from aiida.plugins import WorkflowFactory
from aiida.plugins import CalculationFactory


@click.command('inpgen')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.SETTINGS()
@options.DAEMON()
@options.OPTION_NODE()
@options.QUEUE_NAME()
def launch_inpgen(structure, inpgen, calc_parameters, settings, daemon, option_node, queue):
    """
    Launch an inpgen calcjob on given input

    If no code is given it queries the DB for inpgen codes and uses the one with
    the newest creation time.

    Either structure or anysource_structure can be specified.
    Default structure is Si bulk.
    """

    options_d = {
        'withmpi': False,
        'max_wallclock_seconds': 6000,
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'queue_name': queue
    }
    if option_node:
        opt_dict = option_node.get_dict()
        for key, val in opt_dict.items():
            options_d[key] = val

    process_class = CalculationFactory('fleur.inpgen')
    inputs = {
        'code': inpgen,
        'structure': structure,
        'parameters': calc_parameters,
        'settings': settings,
        'metadata': {
            'options': options_d
        }
    }

    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('fleur')
@options.FLEURINP()
@options.FLEUR()
@options.REMOTE()
@options.SETTINGS()
@options.DAEMON()
@options.MAX_NUM_MACHINES()
@options.MAX_WALLCLOCK_SECONDS()
@options.NUM_MPIPROCS_PER_MACHINE()
@options.OPTION_NODE()
@options.WITH_MPI()
@options.QUEUE_NAME()
@click.option('--launch_base/--no-launch_base',
              is_flag=True,
              default=True,
              show_default=True,
              help=('Run the base_fleur workchain, which also handles errors instead '
                    'of a single fleur calcjob.'))
def launch_fleur(fleurinp, fleur, parent_folder, settings, daemon, max_num_machines, max_wallclock_seconds,
                 num_mpiprocs_per_machine, option_node, with_mpi, launch_base, queue):
    """
    Launch a base_fleur workchain.
    If launch_base is False launch a single fleur calcjob instead.

    """

    process_class = CalculationFactory('fleur.fleur')
    workchain_class = WorkflowFactory('fleur.base')

    options_d = {
        'withmpi': with_mpi,
        'queue_name': queue,
        'max_wallclock_seconds': max_wallclock_seconds,
        'resources': {
            'num_machines': max_num_machines,
            'num_mpiprocs_per_machine': num_mpiprocs_per_machine,
        }
    }
    if option_node:
        opt_dict = option_node.get_dict()
        for key, val in opt_dict.items():
            options_d[key] = val

    inputs = {
        'code': fleur,
        'fleurinpdata': fleurinp,
        'parent_folder': parent_folder,
        'settings': settings,
        'metadata': {
            'options': options_d
        }
    }

    if not launch_base:
        inputs = clean_nones(inputs)
        builder = process_class.get_builder()
        builder.update(inputs)
    else:
        if option_node is None:
            option_node = Dict(
                dict={
                    'withmpi': with_mpi,
                    'max_wallclock_seconds': max_wallclock_seconds,
                    'resources': {
                        'num_machines': max_num_machines,
                        'num_mpiprocs_per_machine': num_mpiprocs_per_machine
                    }
                })

        inputs_base = {
            'code': fleur,
            'fleurinpdata': fleurinp,
            'parent_folder': parent_folder,
            'settings': settings,
            'options': option_node
        }
        inputs_base = clean_nones(inputs_base)
        builder = workchain_class.get_builder()
        builder.update(**inputs_base)

    utils.launch_process(builder, daemon)


@click.command('scf')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.SETTINGS()
@options.FLEURINP()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.REMOTE()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_scf(structure, inpgen, calc_parameters, fleurinp, fleur, wf_parameters, parent_folder, daemon, settings,
               option_node):
    """
    Launch a scf workchain
    """
    workchain_class = WorkflowFactory('fleur.scf')
    inputs = {
        'inpgen': inpgen,
        'fleur': fleur,
        'structure': structure,
        'fleurinp': fleurinp,
        'wf_parameters': wf_parameters,
        'calc_parameters': calc_parameters,
        'remote_data': parent_folder,
        'settings': settings,
        'options': option_node
    }

    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('relax')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.SCF_PARAMETERS()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_relax(structure, inpgen, calc_parameters, fleur, wf_parameters, scf_parameters, daemon, settings,
                 option_node):
    """
    Launch a base relax workchain

    # TODO final scf input
    """
    workchain_class = WorkflowFactory('fleur.base_relax')
    inputs = {
        'scf': {
            'wf_parameters': scf_parameters,
            'structure': structure,
            'calc_parameters': calc_parameters,
            'options': option_node,
            'inpgen': inpgen,
            'fleur': fleur
        },
        'wf_parameters': wf_parameters
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('eos')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.SCF_PARAMETERS()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_eos(structure, inpgen, calc_parameters, fleur, wf_parameters, scf_parameters, daemon, settings, option_node):
    """
    Launch a eos workchain
    """
    workchain_class = WorkflowFactory('fleur.eos')
    inputs = {
        'scf': {
            'wf_parameters': scf_parameters,
            'calc_parameters': calc_parameters,
            'options': option_node,
            'inpgen': inpgen,
            'fleur': fleur
        },
        'wf_parameters': wf_parameters,
        'structure': structure
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('banddos')
@options.FLEURINP()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.REMOTE()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_banddos(fleurinp, fleur, wf_parameters, parent_folder, daemon, settings, option_node):
    """
    Launch a banddos workchain
    """
    workchain_class = WorkflowFactory('fleur.banddos')
    inputs = {
        'wf_parameters': wf_parameters,
        'fleur': fleur,
        'remote': parent_folder,
        'fleurinp': fleurinp,
        'options': option_node
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('init_cls')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEURINP()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_init_cls(structure, inpgen, calc_parameters, fleurinp, fleur, wf_parameters, daemon, settings, option_node):
    """
    Launch an init_cls workchain
    """
    workchain_class = WorkflowFactory('fleur.init_cls')
    inputs = {
        'calc_parameters': calc_parameters,
        'options': option_node,
        'inpgen': inpgen,
        'fleur': fleur,
        'wf_parameters': wf_parameters,
        'structure': structure
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('corehole')
@options.STRUCTURE_OR_FILE(default=defaults.get_si_bulk_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEURINP()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_corehole(structure, inpgen, calc_parameters, fleurinp, fleur, wf_parameters, daemon, settings, option_node):
    """
    Launch a corehole workchain
    """
    workchain_class = WorkflowFactory('fleur.corehole')
    inputs = {
        'calc_parameters': calc_parameters,
        'options': option_node,
        'inpgen': inpgen,
        'fleur': fleur,
        'wf_parameters': wf_parameters,
        'structure': structure
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('mae')
@options.STRUCTURE_OR_FILE(default=defaults.get_fept_film_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEURINP()
@options.FLEUR()
@options.WF_PARAMETERS()
@options.SCF_PARAMETERS()
@options.REMOTE()
@options.DAEMON()
@options.SETTINGS()
@options.OPTION_NODE()
def launch_mae(structure, inpgen, calc_parameters, fleurinp, fleur, wf_parameters, scf_parameters, parent_folder,
               daemon, settings, option_node):
    """
    Launch a mae workchain
    """
    workchain_class = WorkflowFactory('fleur.mae')
    inputs = {
        'scf': {
            'wf_parameters': scf_parameters,
            'structure': structure,
            'calc_parameters': calc_parameters,
            'settings': settings,
            'inpgen': inpgen,
            'fleur': fleur
        },
        'wf_parameters': wf_parameters,
        'fleurinp': fleurinp,
        'remote': parent_folder,
        'fleur': fleur,
        'options': option_node
    }

    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('create_magnetic')
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEUR()
@options.WF_PARAMETERS(required=True)
@options.EOS_PARAMETERS()
@options.SCF_PARAMETERS()
@options.RELAX_PARAMETERS()
@options.DAEMON()
@options.OPTION_NODE()
def launch_create_magnetic(inpgen, calc_parameters, fleur, wf_parameters, eos_parameters, scf_parameters,
                           relax_parameters, daemon, option_node):
    """
    Launch a create_magnetic workchain
    """
    workchain_class = WorkflowFactory('fleur.create_magnetic')
    inputs = {
        'eos': {
            'scf': {
                'wf_parameters': scf_parameters,
                'calc_parameters': calc_parameters,
                'options': option_node,
                'inpgen': inpgen,
                'fleur': fleur
            },
            'wf_parameters': eos_parameters
        },
        'relax': {
            'scf': {
                'wf_parameters': scf_parameters,
                'calc_parameters': calc_parameters,
                'options': option_node,
                'inpgen': inpgen,
                'fleur': fleur
            },
            'wf_parameters': relax_parameters,
            'label': 'relaxation',
        },
        'wf_parameters': wf_parameters
    }

    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('ssdisp')
@options.STRUCTURE_OR_FILE(default=defaults.get_fept_film_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEUR()
@options.WF_PARAMETERS(required=True)
@options.SCF_PARAMETERS()
@options.DAEMON()
@options.OPTION_NODE()
def launch_ssdisp(structure, inpgen, calc_parameters, fleur, wf_parameters, scf_parameters, daemon, option_node):
    """
    Launch a ssdisp workchain
    """
    workchain_class = WorkflowFactory('fleur.ssdisp')
    inputs = {
        'scf': {
            'wf_parameters': scf_parameters,
            'structure': structure,
            'calc_parameters': calc_parameters,
            'options': option_node,
            'inpgen': inpgen,
            'fleur': fleur
        },
        'wf_parameters': wf_parameters,
        'fleur': fleur,
        'options': option_node
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)


@click.command('dmi')
@options.STRUCTURE_OR_FILE(default=defaults.get_fept_film_structure, show_default=True)
@options.INPGEN()
@options.CALC_PARAMETERS()
@options.FLEUR()
@options.WF_PARAMETERS(required=True)
@options.SCF_PARAMETERS()
@options.DAEMON()
@options.OPTION_NODE()
def launch_dmi(structure, inpgen, calc_parameters, fleur, wf_parameters, scf_parameters, daemon, option_node):
    """
    Launch a dmi workchain
    """
    click.echo('Not implemented yet, sorry. Please implement me!')
    workchain_class = WorkflowFactory('fleur.dmi')
    inputs = {
        'scf': {
            'wf_parameters': scf_parameters,
            'structure': structure,
            'calc_parameters': calc_parameters,
            'options': option_node,
            'inpgen': inpgen,
            'fleur': fleur
        },
        'wf_parameters': wf_parameters,
        'fleur': fleur,
        'options': option_node
    }
    inputs = clean_nones(inputs)
    builder = workchain_class.get_builder()
    builder.update(inputs)
    utils.launch_process(builder, daemon)



# -*- coding: utf-8 -*-
'method' : 'valence',      # what method to use, default for valence to highest open shell
'hole_charge' : 1.0,       # what is the charge of the corehole? 0<1.0
'atoms' : ['all'],         # coreholes on what atoms, positions or index for list,
                           # or element ['Be', (0.0, 0.5, 0.334), 3]
'corelevel': ['all'],      # coreholes on which corelevels [ 'Be1s', 'W4f', 'Oall'...]
'supercell_size' : [2,1,1],# size of the supercell [nx,ny,nz]
'para_group' : None,       # use parameter nodes from a parameter group
'relax' : False,           # relax the unit cell first?
'relax_mode': 'Fleur',     # what releaxation do you want
'relax_para' : 'default',  # parameter dict for the relaxation
'scf_para' : 'default',    # wf parameter dict for the scfs
'same_para' : True,        # enforce the same atom parameter/cutoffs on the corehole calc and ref
'magnetic' : True          # jspins=2, makes a difference for coreholes



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module you find the workflow 'fleur_optimize_parameter_wc', which finds
working/(in the future, optiomal) flapw parameters for a given Structure
"""

#import numpy as np

from aiida.plugins import DataFactory
from aiida.orm import Code, load_node
from aiida.orm import StructureData, RemoteData, Dict
from aiida.engine import WorkChain, ToContext  #,Outputs
from aiida.engine import submit
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur, get_inputs_inpgen
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.data.fleurinp import FleurinpData


class fleur_optimize_parameters_wc(WorkChain):
    """
    This workflow finds out working/(in the future optimal) flapw parameters
    from a structure. For now, it runs inpgen on the structure and uses
    the Fleur defaults.

    :param wf_parameters: Dict node, optional, protocol specification
                          will be parsed like this to fleur_eos_wc
    :param structure: StructureData node, bulk crystal structure
    :param inpgen: Code node,
    :param fleur: Code node,

    :return output_optimized_wc_para: Dict node, contains relevant output information
                                      about general succces.
    :return optimized_para: Dict node usable by inpgen
    :return optimized_fleurinp: FleurinpData with optimized

    """

    _workflowversion = '0.1.0'

    _default_wf_para = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(
                dict={
                    'resources': {
                        'num_machines': 1
                    },  #, "num_mpiprocs_per_machine" : 12},
                    'walltime_sec': 60 * 60,
                    'queue_name': '',
                    'custom_scheduler_commands': ''
                }))
        spec.input('structure', valid_type=StructureData, required=True)
        spec.input('inpgen', valid_type=Code, required=True)
        spec.input('fleur', valid_type=Code, required=False)
        spec.outline(cls.start, cls.run_inpgen, cls.determine_parameters, cls.return_results)
        #spec.dynamic_output()

    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report(f'started fleur_optimize_parameter workflow version {self._workflowversion}')
        self.report('Workchain node identifiers: ')  #{}".format(ProcessRegistry().current_calc_node))

        ### input check ###

        # initialize contexts

        self.ctx.successful = True
        # Check on inputnodes

        inputs = self.inputs

        # wf_parameters:
        wf_dict = inputs.wf_parameters.get_dict()

        # set values, or DEFAULTS
        self.ctx.serial = wf_dict.get('serial', False)
        self.ctx.custom_scheduler_commands = wf_dict.get('custom_scheduler_commands', '')
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)
        self.ctx.description_wf = self.inputs.get('_description', '') + '|fleur_optimize_parameters_wc|'
        self.ctx.walltime_sec = wf_dict.get('walltime_sec', 60 * 60)
        self.ctx.resources = wf_dict.get('resources', {'num_machines': 1})
        self.ctx.queue = wf_dict.get('queue_name', '')

        # codes
        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for inpgen of FLEUR does not use the plugin fleur.inpgen')
                self.control_end_wc(error)
                self.abort(error)

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                self.control_end_wc(error)
                self.abort(error)

    def run_inpgen(self):
        """
        So far run inpgen and see what you get
        """

        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        label = 'scf: inpgen'
        description = f'{self.ctx.description_wf} inpgen on {self.ctx.formula}'

        inpgencode = self.inputs.inpgen
        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None

        options = {
            'max_wallclock_seconds': self.ctx.walltime_sec,
            'resources': self.ctx.resources,
            'queue_name': self.ctx.queue
        }

        inputs = get_inputs_inpgen(structure, inpgencode, options, label, description, params=params)
        self.report('INFO: run inpgen')
        future = submit(FleurinputgenCalculation, **inputs)

        return ToContext(inpgen=future, last_calc=future)

    def determine_parameters(self):
        """
        Determine_parameters, so far extract them from FleurinpData
        """

        try:
            fleurin = self.ctx['inpgen'].outputs.fleurinpData
        except AttributeError:
            error = 'No fleurinpData found, inpgen failed'
            self.control_end_wc(error)

        # TODO this has to be implemented
        self.ctx.optimal_parameters = fleurin.get_parameterdata(fleurin)

    # TODO determine optimasation
    # look at number of atoms
    # basis size nvd not known, only after fleur run
    # can this be examined via kmax?
    # number of kpoints
    #

    def return_results(self):
        """
        Prepare and return the result nodes
        """

        if self.ctx.optimal_parameters:
            optimal_parameters_uuid = self.ctx.optimal_parameters.uuid
        else:
            optimal_parameters_uuid = None

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            'structure': self.inputs.structure.uuid,
            'optimal_para': optimal_parameters_uuid,
            'successful': self.ctx.successful
        }

        outnode = Dict(dict=out)

        returndict = {}

        if self.ctx.successful:
            self.report('Done, fleur_optimize_parameter_wc calculation complete')

            returndict['output_optimized_wc_para'] = outnode
            returndict['optimized_para'] = self.ctx.optimal_parameters
            #optimized_fleurinp
        else:
            self.report('Done, but something failed in fleur_optimize_parameter_wc.')

        # create link to workchain node
        for link_name, node in returndict.items():
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain. Can initalize the output nodes,
        or set things in the contents
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg)  # because return_results still fails somewhen
        self.return_results()
        #self.abort_nowait(errormsg)
        self.abort(errormsg)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Calculation function to compute a k-point mesh for a structure with a
guaranteed minimum k-point distance.
"""
from aiida.engine import calcfunction
from .merge_parameter import merge_parameter


@calcfunction
def create_kpoints_from_distance_parameter(structure, cf_para, calc_parameters=None):
    """
    Generate a uniformly spaced kpoint mesh for a given structure
    and merge it into a given calc_parameter node or create a new one

    The spacing between kpoints in reciprocal space is guaranteed to be at least the defined distance.
    :param structure: the StructureData to which the mesh should apply
    :param cf_para: an AiiDA Dict which has all the keys inside needed for the
        create_kpoints_from_distance_parameter_ncf call:
        distance: a float with the desired distance between kpoints in reciprocal space
        force_even: a Bool to specify whether the generated mesh should have only
                       even values in direction of periodic boundary conditions
        force_odd: a Bool to specify whether the generated mesh should have only odd values
        force_parity: a bool to specify whether the generated mesh should maintain parity
    :param calc_parameters: a Dict which contains calc parameters for inpgen
    :returns: Dict node with the generated mesh, merged with given calc_parameters
    """
    cf_dict = cf_para.get_dict()
    distance = cf_dict.get('distance', 0.1)
    force_parity = cf_dict.get('force_parity', False)
    force_odd = cf_dict.get('force_odd', False)
    force_even = cf_dict.get('force_even', False)
    include_gamma = cf_dict.get('include_gamma', False)
    # we could also parse directly the dict to the function with ** but this way we ignore
    # wrong or additional keys.
    new_calc_para = create_kpoints_from_distance_parameter_ncf(structure,
                                                               distance=distance,
                                                               force_parity=force_parity,
                                                               force_odd=force_odd,
                                                               force_even=force_even,
                                                               include_gamma=include_gamma,
                                                               calc_parameters=calc_parameters)
    return new_calc_para


def create_kpoints_from_distance_parameter_ncf(structure, distance, force_parity, force_odd=False, \
                                               force_even=False,include_gamma=False, calc_parameters=None):
    """
    Generate a uniformly spaced kpoint mesh for a given structure
    and merge it into a given calc_parameter node or create a new one.
    Does not keep the provenance

    The spacing between kpoints in reciprocal space is guaranteed to be at least the defined distance.
    :param structure: the StructureData to which the mesh should apply
    :param distance: a Float with the desired distance between kpoints in reciprocal space
    :param force_parity: a Bool to specify whether the generated mesh should maintain parity
    :param force_even: a Bool to specify whether the generated mesh should have only
                       even values in direction of periodic boundary conditions
    :param force_odd: a Bool to specify whether the generated mesh should have only odd values
    :param calc_parameters: a Dict which contains calc parameters for inpgen
    :returns: Dict node with the generated mesh, merged with given calc_parameters
    :returns: Dict node with the generated mesh
    """
    import numpy
    from numpy import linalg
    from aiida.orm import Dict

    epsilon = 1E-5

    the_cell = structure.cell
    reciprocal_cell = 2. * numpy.pi * numpy.linalg.inv(numpy.array(the_cell)).transpose()

    kpointsmesh = [
        max(int(numpy.ceil(round(numpy.linalg.norm(b) / distance, 5))), 1) if pbc else 1
        for pbc, b in zip(structure.pbc, reciprocal_cell)
    ]

    if force_parity:
        kpointsmesh = [k + (k % 2) if pbc else 1 for pbc, k in zip(structure.pbc, kpointsmesh)]

    if force_even:  # force even only applies to directions with non periodic boundary conditions
        kpointsmesh = [k if (k % 2 == 0) else k + 1 for k in kpointsmesh]
        kpointsmesh = [k if pbc else 1 for pbc, k in zip(structure.pbc, kpointsmesh)]

    if force_odd:
        kpointsmesh = [k if (k % 2 == 1) else k + 1 for k in kpointsmesh]
        kpointsmesh = [k if pbc else 1 for pbc, k in zip(structure.pbc, kpointsmesh)]

    lengths_vector = [linalg.norm(vector) for vector in the_cell]

    is_symmetric_cell = all(abs(length - lengths_vector[0]) < epsilon for length in lengths_vector)
    is_symmetric_mesh = all(length == kpointsmesh[0] for length in kpointsmesh)

    # If the vectors of the cell all have the same length, the kpoint mesh should be isotropic as well
    if is_symmetric_cell and not is_symmetric_mesh:
        nkpoints = max(kpointsmesh)
        kpointsmesh = [nkpoints, nkpoints, nkpoints]

    mesh_spec = {'kpt': {'div1': kpointsmesh[0], 'div2': kpointsmesh[1], 'div3': kpointsmesh[2]}}
    if include_gamma:
        mesh_spec['kpt']['gamma'] = True

    new_calc_para = Dict(dict=mesh_spec)
    if calc_parameters is not None:
        # Override false, since we want to keep other kpts keys in calc_parameters
        new_calc_para = merge_parameter(new_calc_para, calc_parameters, overwrite=False, merge=True)

    return new_calc_para



# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit
from aiida.plugins import WorkflowFactory


fleur_init_cls_wc = WorkflowFactory('fleur.initial_cls')
struc = load_node(STRUCTURE_PK)
flapw_para = load_node(PARAMETERS_PK)
fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)

options = Dict(dict={'resources': {'num_machines': 2, 'num_mpiprocs_per_machine': 24},
                         'queue_name': '',
                         'custom_scheduler_commands': '',
                         'max_wallclock_seconds':  60*60})

wf_para_initial = Dict(dict={
  'references': {'Be': '257d8ae8-32b3-4c95-8891-d5f527b80008',
                 'W': 'c12c999c-9a00-4866-b6ef-9bb5d28e7797'},
  'scf_para': {'density_criterion': 5e-06, 'fleur_runmax': 3, 'itmax_per_run': 80}})

# launch workflow
initial_res = submit(fleur_init_cls_wc, wf_parameters=wf_para_initial, structure=struc,
             calc_parameters=flapw_para, options=options, fleur=fleur, inpgen=inpgen,
             label='test initial cls', description='fleur_initial_cls test')



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Options commonly used throughout the aiida-fleur command line interface.
To standardize at least of the options, they are kept as close as possible to
aiida-core and aiida-quantumespresso
"""
import click
from aiida.cmdline.params import types
from aiida.cmdline.params.options import OverridableOption
from .defaults import get_inpgen, get_fleur, get_si_bulk_structure
from .types import StructureNodeOrFileParamType

STRUCTURE_OR_FILE = OverridableOption(
    '-s',
    '--structure',
    type=StructureNodeOrFileParamType(),
    help='StructureData node, given by pk or uuid or file in any for mat which will be converted.')

STRUCTURE = OverridableOption('-s',
                              '--structure',
                              type=types.DataParamType(sub_classes=('aiida.data:structure',)),
                              help='StructureData node, given by pk or uuid.')

FULL_PROVENANCE = OverridableOption('-fp',
                                    '--full-provenance',
                                    is_flag=True,
                                    default=False,
                                    show_default=True,
                                    help=('Store the full or reduced provenance. Example with the "-as" '
                                          'also the given file will be stored in the database together with a '
                                          'calcfunction extracting the structure.'))

INPGEN = OverridableOption('-i',
                           '--inpgen',
                           type=types.CodeParamType(entry_point='fleur.inpgen'),
                           default=get_inpgen,
                           show_default=True,
                           help='A code node or label for an inpgen executable.')

FLEUR = OverridableOption('-f',
                          '--fleur',
                          type=types.CodeParamType(entry_point='fleur.fleur'),
                          default=get_fleur,
                          show_default=True,
                          help='A code node or label for a fleur executable.')

FLEURINP = OverridableOption('-inp',
                             '--fleurinp',
                             type=types.DataParamType(sub_classes=('aiida.data:fleur.fleurinp',)),
                             help='FleurinpData node for the fleur calculation.')

CALC_PARAMETERS = OverridableOption(
    '-calc_p',
    '--calc-parameters',
    type=types.DataParamType(sub_classes=('aiida.data:dict',)),
    help='Dict with calculation (FLAPW) parameters to build, which will be given to inpgen.')

SETTINGS = OverridableOption('-set',
                             '--settings',
                             type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                             help='Settings node for the calcjob.')

WF_PARAMETERS = OverridableOption('-wf',
                                  '--wf-parameters',
                                  type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                                  help='Dict containing parameters given to the workchain.')

SCF_PARAMETERS = OverridableOption('-scf',
                                   '--scf-parameters',
                                   type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                                   help='Dict containing parameters given to the sub SCF workchains.')

EOS_PARAMETERS = OverridableOption('-eos',
                                   '--eos-parameters',
                                   type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                                   help='Dict containing wf parameters given to the sub EOS workchains.')

RELAX_PARAMETERS = OverridableOption('-relax',
                                     '--relax-parameters',
                                     type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                                     help='Dict containing wf parameters given to the sub relax workchains.')

OPTION_NODE = OverridableOption('-opt',
                                '--option-node',
                                type=types.DataParamType(sub_classes=('aiida.data:dict',)),
                                help='Dict, an option node for the workchain.')

MAX_NUM_MACHINES = OverridableOption('-N',
                                     '--max-num-machines',
                                     type=click.INT,
                                     default=1,
                                     show_default=True,
                                     help='The maximum number of machines (nodes) to use for the calculations.')

MAX_WALLCLOCK_SECONDS = OverridableOption('-W',
                                          '--max-wallclock-seconds',
                                          type=click.INT,
                                          default=1800,
                                          show_default=True,
                                          help='The maximum wallclock time in seconds to set for the calculations.')

QUEUE_NAME = OverridableOption('-q',
                               '--queue',
                               type=click.STRING,
                               default='',
                               show_default=True,
                               help='The queue name to submit to.')

NUM_MPIPROCS_PER_MACHINE = OverridableOption('-M',
                                             '--num-mpiprocs-per-machine',
                                             type=click.INT,
                                             default=2,
                                             show_default=True,
                                             help='Run the simulation with so many num-mpi-procs-per-machine.')

WITH_MPI = OverridableOption('-I',
                             '--with-mpi',
                             is_flag=True,
                             default=False,
                             show_default=True,
                             help='Run the calculations with MPI enabled.')

REMOTE = OverridableOption('-P',
                           '--parent-folder',
                           'parent_folder',
                           type=types.DataParamType(sub_classes=('aiida.data:remote',)),
                           show_default=True,
                           required=False,
                           help='The PK of a parent remote folder (for restarts).')

DAEMON = OverridableOption('-d',
                           '--daemon',
                           is_flag=True,
                           default=False,
                           show_default=True,
                           help='Submit the process to the daemon instead of running it locally.')

CLEAN_WORKDIR = OverridableOption(
    '-x',
    '--clean-workdir',
    is_flag=True,
    default=False,
    show_default=True,
    help='Clean the remote folder of all the launched calculations after completion of the workchain.')

SHOW = OverridableOption('--show/--no-show',
                         default=True,
                         show_default=True,
                         help='Show the main output of the command.')



"""
This module defines XML modifying functions, that require an aiida node as input
"""


def set_kpointsdata_f(xmltree, schema_dict, kpointsdata_uuid, name=None, switch=False, kpoint_type='path'):
    """This function creates a kpoint list in the inp.xml from a :py:class:`~aiida.orm.KpointsData` Node
    If no weights are given the weight is distibuted equally along the kpoints

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param kpointsdata_uuid: node identifier or :class:`~aiida.orm.KpointsData` node to be written into ``inp.xml``
    :param name: str name to give the newly entered kpoint list (only MaX5 or later)
    :param switch: bool if True the entered kpoint list will be used directly (only Max5 or later)
    :param kpoint_type: str of the type of kpoint list given (mesh, path, etc.) only Max5 or later

    :return: xmltree with entered kpoint list
    """
    # TODO: check on weights,
    import numpy as np
    from aiida.orm import KpointsData, load_node
    from aiida.common.exceptions import InputValidationError
    from masci_tools.util.xml.xml_setters_names import set_kpointlist

    if not isinstance(kpointsdata_uuid, KpointsData):
        KpointsDataNode = load_node(kpointsdata_uuid)
    else:
        KpointsDataNode = kpointsdata_uuid

    if not isinstance(KpointsDataNode, KpointsData):
        raise InputValidationError('The node given is not a valid KpointsData node.')

    try:
        kpoints, weights = KpointsDataNode.get_kpoints(also_weights=True, cartesian=False)
    except AttributeError:
        kpoints = KpointsDataNode.get_kpoints(cartesian=False)
        weights = np.ones(len(kpoints)) / len(kpoints)

    labels = KpointsDataNode.labels

    labels_dict = None
    if labels is not None:
        labels_dict = dict(labels)

    # try:
    #     KpointsDataNode.get_kpoints_mesh()
    #     kpoint_type = 'mesh'
    # except AttributeError:
    #     kpoint_type = 'path'

    if schema_dict.inp_version <= (0, 31):
        xmltree = set_kpointlist(xmltree, schema_dict, kpoints, weights)
    else:
        xmltree = set_kpointlist(xmltree,
                                 schema_dict,
                                 kpoints,
                                 weights,
                                 special_labels=labels_dict,
                                 kpoint_type=kpoint_type,
                                 name=name,
                                 switch=switch)

    return xmltree


FLEURINPMODIFIER_EXTRA_FUNCS = {'schema_dict': {'set_kpointsdata': set_kpointsdata_f}}



'''
This module collects helper methods which are commonly used for dictionaries.
'''
from collections.abc import Mapping

from aiida.common.extendeddicts import AttributeDict
from aiida.orm.nodes.data.dict import Dict
# FIXME, there might be overlap with tools/dict_utils or other tools.


def update_mapping(original, source):
    """
    Update a nested dictionary with another optionally nested dictionary. The dictionaries may be
    plain Mapping objects or Dict nodes. If the original dictionary is an instance of Dict
    the returned dictionary will also be wrapped in Dict.

    :param original: Mapping object or Dict instance
    :param source: Mapping object or Dict instance
    :return: the original dictionary updated with the source dictionary
    """
    return_parameter_data = False

    if isinstance(original, Dict):
        return_parameter_data = True
        original = original.get_dict()

    if isinstance(source, Dict):
        source = source.get_dict()

    for key, value in source.items():
        if key in original and isinstance(value, (Dict, Mapping)) and isinstance(original[key], (Dict, Mapping)):
            original[key] = update_mapping(original[key], value)
        else:
            original[key] = value

    if return_parameter_data:
        original = Dict(dict=original)

    return original


def prepare_process_inputs(process, inputs):
    """
    Prepare the inputs for submission for the given process, according to its spec. That is to say
    that when an input is found in the inputs that corresponds to an input port in the spec of the
    process that expects a Dict, yet the value in the inputs is a plain dictionary, the value will
    be wrapped in by the Dict class to create a valid input.

    :param process: sub class of Process for which to prepare the inputs dictionary
    :param inputs: a dictionary of inputs intended for submission of the process
    :return: a dictionary with all bare dictionaries wrapped in Dict if dictated by process spec
    """
    prepared_inputs = AttributeDict()

    try:
        process_spec = process.spec()
    except AttributeError as exc:
        raise ValueError('Process {} does not have a spec') from exc

    for key, value in inputs.items():

        if key not in process_spec.inputs:
            continue

        if process_spec.inputs[key].valid_type == Dict and isinstance(value, dict):
            prepared_inputs[key] = Dict(dict=value)
        else:
            prepared_inputs[key] = value

    return prepared_inputs



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the parser for a FLEUR calculation and methods for parsing
different files produced by FLEUR.

Please implement file parsing routines that they can be executed from outside
the parser. Makes testing and portability easier.
"""
# TODO: warnings
import re
import json
from lxml import etree

from aiida.parsers import Parser
from aiida.orm import Dict
from aiida.common.exceptions import NotExistent

from masci_tools.io.parsers.fleur import outxml_parser
from masci_tools.io.parsers.fleur_schema import InputSchemaDict


class FleurParser(Parser):
    """
    This class is the implementation of the Parser class for FLEUR.
    It parses the FLEUR output if the calculation was successful,
    i.e checks if all files are there that should be and their condition.
    Then it parses the out.xml file and returns a (simple) parameterData node
    with the results of the last iteration.
    Other files (DOS.x, bands.x, relax.xml, ...) are also parsed if they are retrieved.
    """

    _setting_key = 'parser_options'

    def get_linkname_outparams_complex(self):
        """
        Returns the name of the link to the output_complex
        Node contains the Fleur output in a rather complex dictionary.
        """
        return 'output_complex'

    def get_linkname_outparams(self):
        """
        Returns the name of the link to the output_complex
        Node contains the Fleur output in a rather complex dictionary.
        """
        return 'output_parameters'

    def parse(self, **kwargs):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here. Checks presents of files.
        Calls routines to parse them and returns parameter nodes and success.

        :return successful: Bool, if overall parsing was successful or not
        :return new_nodes_list: list of tuples of two (linkname, Dataobject),
                                nodes to be stored by AiiDA

        """

        ####### init some variables ######

        # these files should be at least present after success of a Fleur run
        calc = self.node
        FleurCalculation = calc.process_class

        # this files should be retrieved
        should_retrieve = calc.get_attribute('retrieve_list')

        has_xml_outfile = False
        has_relax_file = False

        ######### Check presence of files ######

        # select the folder object
        # Check that the retrieved folder is there
        try:
            output_folder = self.retrieved
        except NotExistent:
            self.logger.error('No retrieved folder found')
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = output_folder.list_object_names()
        self.logger.info(f'File list: {list_of_files}')

        # has output xml file, otherwise error
        if FleurCalculation._OUTXML_FILE_NAME not in list_of_files:
            self.logger.error(f"XML out not found '{FleurCalculation._OUTXML_FILE_NAME}'")
            has_xml_outfile = False  #Return after the error lines were processed
        else:
            has_xml_outfile = True

        # check if all files expected are there for the calculation
        for file in should_retrieve:
            if file not in list_of_files:
                self.logger.warning(
                    f"Expected file '{file}' not found in retrieved folder, it was probably not created by fleur")

        # check if something was written to the error file
        if FleurCalculation._ERROR_FILE_NAME in list_of_files:
            errorfile = FleurCalculation._ERROR_FILE_NAME
            # read
            try:
                with output_folder.open(errorfile, 'r') as efile:
                    error_file_lines = efile.read()  # Note: read(), not readlines()
            except OSError:
                self.logger.error(f'Failed to open error file: {errorfile}.')
                return self.exit_codes.ERROR_OPENING_OUTPUTS

            if error_file_lines:

                if isinstance(error_file_lines, bytes):
                    error_file_lines = error_file_lines.replace(b'\x00', b' ')
                else:
                    error_file_lines = error_file_lines.replace('\x00', ' ')
                if 'Run finished successfully' not in error_file_lines:
                    self.logger.warning('The following was written into std error and piped to {}'
                                        ' : \n {}'.format(errorfile, error_file_lines))
                    self.logger.error('FLEUR calculation did not finish successfully.')

                    # here we estimate how much memory was available and consumed
                    mpiprocs = self.node.get_attribute('resources').get('num_mpiprocs_per_machine', 1)

                    kb_used = 0.0
                    if has_xml_outfile:
                        with output_folder.open(FleurCalculation._OUTXML_FILE_NAME,
                                                'r') as out_file:  # lazy out.xml parsing
                            outlines = out_file.read()
                            try:
                                line_avail = re.findall(r'<mem memoryPerNode="\d+', outlines)[0]
                                mem_kb_avail = int(re.findall(r'\d+', line_avail)[0])
                            except IndexError:
                                mem_kb_avail = 1.0
                                self.logger.info('Did not manage to find memory available info.')
                            else:
                                usage_json = FleurCalculation._USAGE_FILE_NAME
                                if usage_json in list_of_files:
                                    with output_folder.open(usage_json, 'r') as us_file:
                                        usage = json.load(us_file)
                                    kb_used = usage['data']['VmPeak']
                                else:
                                    try:
                                        line_used = re.findall(r'used.+', error_file_lines)[0]
                                        kb_used = int(re.findall(r'\d+', line_used)[2])
                                    except IndexError:
                                        self.logger.info('Did not manage to find memory usage info.')
                    else:
                        kb_used = 0.0
                        mem_kb_avail = 1.0
                        self.logger.info('Did not manage to find memory available info.')
                        self.logger.info('Did not manage to find memory usage info.')

                    # here we estimate how much walltime was available and consumed
                    try:
                        time_avail_sec = self.node.attributes['last_job_info']['requested_wallclock_time_seconds']
                        time_calculated = self.node.attributes['last_job_info']['wallclock_time_seconds']
                        if 0.97 * time_avail_sec < time_calculated:
                            return self.exit_codes.ERROR_TIME_LIMIT
                    except KeyError:
                        pass

                    if (kb_used * mpiprocs / mem_kb_avail > 0.93 or
                            'cgroup out-of-memory handler' in error_file_lines or 'Out Of Memory' in error_file_lines):
                        return self.exit_codes.ERROR_NOT_ENOUGH_MEMORY
                    elif 'TIME LIMIT' in error_file_lines or 'time limit' in error_file_lines:
                        return self.exit_codes.ERROR_TIME_LIMIT
                    elif 'Atom spills out into vacuum during relaxation' in error_file_lines:
                        return self.exit_codes.ERROR_VACUUM_SPILL_RELAX
                    elif 'Error checking M.T. radii' in error_file_lines:
                        return self.exit_codes.ERROR_MT_RADII
                    elif 'No solver linked for Hubbard 1' in error_file_lines:
                        return self.exit_codes.ERROR_MISSING_DEPENDENCY.format(name='edsolver')
                    elif 'Overlapping MT-spheres during relaxation: ' in error_file_lines:
                        overlap_line = re.findall(r'\S+ +\S+ olap: +\S+', error_file_lines)[0].split()
                        with output_folder.open('relax.xml', 'r') as rlx:
                            schema_dict = InputSchemaDict.fromVersion('0.34')
                            relax_dict = parse_relax_file(rlx, schema_dict)
                            it_number = len(relax_dict['energies']) + 1  # relax.xml was not updated
                        error_params = {
                            'error_name': 'MT_OVERLAP_RELAX',
                            'description': ('This output node contains information'
                                            'about FLEUR error'),
                            'overlapped_indices': overlap_line[:2],
                            'overlaping_value': overlap_line[3],
                            'iteration_number': it_number
                        }
                        link_name = self.get_linkname_outparams()
                        error_params = Dict(dict=error_params)
                        self.out('error_params', error_params)
                        return self.exit_codes.ERROR_MT_RADII_RELAX
                    elif 'parent_folder' in calc.inputs:  # problem in reusing cdn for relaxations, drop cdn
                        if 'fleurinpdata' in calc.inputs:
                            if 'relax.xml' in calc.inputs.fleurinpdata.files:
                                return self.exit_codes.ERROR_DROP_CDN
                        return self.exit_codes.ERROR_FLEUR_CALC_FAILED
                    else:
                        return self.exit_codes.ERROR_FLEUR_CALC_FAILED

        # if a relax.xml was retrieved
        if FleurCalculation._RELAX_FILE_NAME in list_of_files:
            self.logger.info('relax.xml file found in retrieved folder')
            has_relax_file = True

        ####### Parse the files ########

        if not has_xml_outfile:
            return self.exit_codes.ERROR_NO_OUTXML
        # open output file

        with output_folder.open(FleurCalculation._OUTXML_FILE_NAME, 'rb') as outxmlfile_opened:
            success = True
            parser_info = {}
            try:
                out_dict = outxml_parser(outxmlfile_opened, parser_info_out=parser_info, ignore_validation=True)
            except (ValueError, FileNotFoundError, KeyError) as exc:
                self.logger.error(f'XML output parsing failed: {str(exc)}')
                success = False

        # Call routines for output node creation
        if not success:
            self.logger.error('Parsing of XML output file was not successfull.')
            outxml_params = Dict(dict=parser_info)
            link_name = self.get_linkname_outparams()
            self.out(link_name, outxml_params)
            return self.exit_codes.ERROR_XMLOUT_PARSING_FAILED
        elif out_dict:
            outxml_params = Dict(dict={**out_dict, **parser_info})
            link_name = self.get_linkname_outparams()
            self.out(link_name, outxml_params)
        else:
            self.logger.error('Something went wrong, no out_dict found')
            outxml_params = Dict(dict=parser_info)
            link_name = self.get_linkname_outparams()
            self.out(link_name, outxml_params)

        if has_relax_file:
            relax_name = FleurCalculation._RELAX_FILE_NAME
            try:
                fleurinp = calc.inputs.fleurinpdata
            except NotExistent:
                old_relax_text = ''
            else:
                if relax_name in fleurinp.list_object_names():
                    with fleurinp.open(relax_name, 'r') as rlx:
                        old_relax_text = rlx.read()
                else:
                    old_relax_text = ''

            inp_version = outxml_params.get_dict().get('input_file_version', '0.34')
            schema_dict = InputSchemaDict.fromVersion(inp_version)
            # dummy comparison between old and new relax
            with output_folder.open(relax_name, 'rb') as rlx:
                new_relax_text = rlx.read()
                if new_relax_text != old_relax_text:
                    try:
                        relax_dict = parse_relax_file(rlx, schema_dict)
                    except etree.XMLSyntaxError:
                        return self.exit_codes.ERROR_RELAX_PARSING_FAILED
                    self.out('relax_parameters', relax_dict)


def parse_relax_file(relax_file, schema_dict):
    """
    This function parsers relax.xml output file and
    returns a Dict containing all the data given there.
    """
    from masci_tools.util.xml.xml_getters import get_relaxation_information

    relax_file.seek(0)
    tree = etree.parse(relax_file)

    out_dict = get_relaxation_information(tree, schema_dict)

    return Dict(dict=out_dict)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module is the :class:`~aiida_fleur.data.fleurinp.FleurinpData` class, and methods for FLEUR
input manipulation plus methods for extration of AiiDA data structures.
"""
# TODO: maybe add a modify method which returns a fleurinpmodifier class
# TODO: inpxml to dict: maybe kpts should not be written to the dict? same with symmetry
# TODO: test for large input files, I believe the recursion is still quite slow..
# TODO: 2D cell get kpoints and get structure also be carefull with tria = T!!!
# TODO : maybe save when get_structure or get_kpoints was executed on fleurinp,
# because otherwise return this node instead of creating a new one!

import os
import io
import re

from lxml import etree
import warnings

from aiida.orm import Data, Node, load_node, CalcJobNode, Bool
from aiida.common.exceptions import InputValidationError, ValidationError
from aiida.engine.processes.functions import calcfunction as cf


def get_fleurinp_from_folder_data(folder_node, store=False, additional_files=None):
    """
    Create FleurinpData object from the given RemoteData object

    :param remote_node: RemoteData to use for the generation of the FleurinpData
    :param store: bool, if True the FleurinpData object will be stored after generation

    :returns: FleurinpData object with the input xml files from the retrieved folder
              of the calculation associated RemoteData
    """
    if additional_files is None:
        additional_files = []

    input_xml_files = [file for file in folder_node.list_object_names() if file.endswith('.xml') and 'out' not in file]

    fleurinp = FleurinpData(files=input_xml_files + additional_files, node=folder_node)
    if store:
        fleurinp.store()

    return fleurinp


def get_fleurinp_from_remote_data(remote_node, store=False, additional_files=None):
    """
    Create FleurinpData object from the given RemoteData object

    :param remote_node: RemoteData to use for the generation of the FleurinpData
    :param store: bool, if True the FleurinpData object will be stored after generation

    :returns: FleurinpData object with the input xml files from the retrieved folder
              of the calculation associated RemoteData
    """

    for link in remote_node.get_incoming().all():
        if isinstance(link.node, CalcJobNode):
            parent_calc_node = link.node
    retrieved = parent_calc_node.get_outgoing().get_node_by_label('retrieved')

    return get_fleurinp_from_folder_data(retrieved, store=store, additional_files=additional_files)


class FleurinpData(Data):
    """
    AiiDA data object representing everything a FLEUR calculation needs.

    It is initialized with an absolute path to an ``inp.xml`` file or a
    FolderData node containing ``inp.xml``.
    Other files can also be added that will be copied to the remote machine, where the
    calculation takes place.

    It stores the files in the repository and stores the input parameters of the
    ``inp.xml`` file of FLEUR in the database as a python dictionary (as internal attributes).
    When an ``inp.xml`` (name important!) file is added to files, parsed into a XML tree
    and validated against the XML schema file for the given file version. These XML schemas
    are provided by the `masci-tools` library


    FleurinpData also provides the user with
    methods to extract AiiDA StructureData, KpointsData nodes
    and Dict nodes with LAPW parameters.

    Remember that most attributes of AiiDA nodes can not be changed after they
    have been stored in the database! Therefore, you have to use the FleurinpModifier class and its
    methods if you want to change somthing in the ``inp.xml`` file. You will retrieve a new
    FleurinpData that way and start a new calculation from it.
    """

    __version__ = '0.5.0'

    def __init__(self, **kwargs):
        """
        Initialize a FleurinpData object set the files given
        """
        files = kwargs.pop('files', None)
        for filename in files:
            if 'inp.xml' in filename:
                files.pop(files.index(filename))
                files.append(filename)
                break
        node = kwargs.pop('node', None)
        super().__init__(**kwargs)

        if files:
            if node:
                self.set_files(files, node=node)
            else:
                self.set_files(files)

    @property
    def parser_info(self):
        """
        Dict property, with the info and warnings from the inpxml_parser
        """
        return self.get_extra('_parser_info', {})

    @parser_info.setter
    def parser_info(self, info_dict):
        """
        Setter for has_schema
        """
        self.set_extra('_parser_info', info_dict)

    # files
    @property
    def files(self):
        """
        Returns the list of the names of the files stored
        """
        return self.get_attribute('files', [])

    @files.setter
    def files(self, filelist, node=None):
        """
        Add a list of files to FleurinpData.
        Alternative use setter method.

        :param files: list of filepaths or filenames of node is specified
        :param node: a Folder node containing files from the filelist
        """
        for file1 in filelist:
            self.set_file(file1, node=node)

    def set_files(self, files, node=None):
        """
        Add the list of files to the :class:`~aiida_fleur.data.fleurinp.FleurinpData` instance.
        Can by used as an alternative to the setter.

        :param files: list of abolute filepaths or filenames of node is specified
        :param node: a :class:`~aiida.orm.FolderData` node containing files from the filelist
        """
        for file1 in files:
            self.set_file(file1, node=node)

    def set_file(self, filename, dst_filename=None, node=None):
        """
        Add a file to the :class:`~aiida_fleur.data.fleurinp.FleurinpData` instance.

        :param filename: absolute path to the file or a filename of node is specified
        :param node: a :class:`~aiida.orm.FolderData` node containing the file
        """
        self._add_path(filename, dst_filename=dst_filename, node=node)

    def open(self, path='inp.xml', mode='r', key=None):  #pylint: disable=arguments-differ
        """
        Returns an open file handle to the content of this data node.

        :param key: name of the file to be opened
        :param mode: the mode with which to open the file handle
        :returns: A file handle in read mode
         """

        if key is not None:
            path = key

        return super().open(path, mode=mode)

    def get_content(self, filename='inp.xml'):
        """
        Returns the content of the single file stored for this data node.

        :returns: A string of the file content
        """
        with self.open(path=filename, mode='r') as handle:
            return handle.read()

    def del_file(self, filename):
        """
        Remove a file from FleurinpData instancefind

        :param filename: name of the file to be removed from FleurinpData instance
        """
        # remove from files attr list
        if filename in self.get_attribute('files'):
            try:
                self.get_attribute('files').remove(filename)
                # self._del_attribute(‘filename')
            except AttributeError:
                # There was no file set
                pass
        # remove from sandbox folder
        if filename in self.list_object_names():  # get_folder_list():
            self.delete_object(filename)

    def _add_path(self, file1, dst_filename=None, node=None):
        """
        Add a single file to the FleurinpData folder.
        The destination name can be different. ``inp.xml`` is a special case.
        file names are stored in the db, the whole file in the reporsitory.

        :param file1: the file to be added, either string, absolute path, or filelike object whose contents to copy
            Hint: Pass io.BytesIO(b"my string") to construct the file directly from a string.
        :param dst_filename: string, new filename for given file in repo and db
        :param node: aiida.orm.Node, usually FolderData if node is given get the 'file1' from the node

        :raise: ValueError, InputValidationError
        :return: None
        """
        # TODO? Maybe only certain files should be allowed to be added
        # contra: has to be maintained, also these files can be inputed from byte strings...
        #_list_of_allowed_files = ['inp.xml', 'enpara', 'cdn1', 'sym.out', 'kpts']

        if node:
            if not isinstance(node, Node):
                node = load_node(node)  # if this fails it will raise
            if file1 not in node.list_object_names():
                # throw error, you try to add something that is not there
                raise ValueError('file1 has to be in the specified node')

            is_filelike = True
            if dst_filename is None:
                final_filename = file1
            else:
                final_filename = dst_filename
            # Override file1 with bytestring of file
            # since we have to use 'with', and node has no method to copy files
            # we read the whole file and write it again later
            # this is not so nice, but we assume that input files are rather small...
            with node.open(file1, mode='rb') as file2:
                file1 = io.BytesIO(file2.read())

        elif isinstance(file1, str):
            is_filelike = False

            if not os.path.isabs(file1):
                file1 = os.path.abspath(file1)
                #raise ValueError("Pass an absolute path for file1: {}".format(file1))

            if not os.path.isfile(file1):
                raise ValueError(f'file1 must exist and must be a single file: {file1}')

            if dst_filename is None:
                final_filename = os.path.split(file1)[1]
            else:
                final_filename = dst_filename
        else:
            is_filelike = True
            if dst_filename is None:
                try:
                    final_filename = os.path.basename(file1.name)  # Not sure if this still works for aiida>2.0
                except AttributeError:
                    final_filename = 'inp.xml'  # fall back to default
            else:
                final_filename = dst_filename

        key = final_filename

        old_file_list = self.list_object_names()
        old_files_list = self.get_attribute('files', [])

        # remove file from folder first if it exists
        if final_filename not in old_file_list:
            old_files_list.append(final_filename)
        else:
            try:
                old_file_list.remove(final_filename)
            except ValueError:
                pass

        if is_filelike:
            try:
                self.put_object_from_filelike(file1, key, mode='wb')
            except TypeError:
                self.put_object_from_filelike(file1, key)
        else:
            self.put_object_from_file(file1, key)

        self.set_attribute('files', old_files_list)  # We want to keep the other files

        ### Special case: 'inp.xml' ###
        # here this is hardcoded, might want to change? get filename from elsewhere

        if final_filename == 'inp.xml':
            # get input file version number
            inp_version_number = None
            lines = self.get_content(filename=final_filename).split('\n')
            for line in lines:
                if re.search('fleurInputVersion', str(line)):
                    inp_version_number = re.findall(r'\d+.\d+', str(line))[0]
                    break
            if inp_version_number is None:
                raise InputValidationError('No fleurInputVersion number found '
                                           'in given input file: {}. {}'
                                           'Please check if this is a valid fleur input file. '
                                           'It can not be validated and I can not use it. '
                                           ''.format(file1, lines))

            self.set_attribute('inp_version', inp_version_number)
            # finally set inp dict of Fleurinpdata
            self._set_inp_dict()

    def _set_inp_dict(self):
        """
        Sets the inputxml_dict from the ``inp.xml`` file attached to FleurinpData

        1. load ``inp.xml`` file
        2. insert all files to include into the etree
        3. call masci-tools input file parser (Validation happens inside here)
        4. set inputxml_dict
        """
        from masci_tools.io.parsers.fleur import inpxml_parser

        #The schema_dict is not needed outside the inpxml_parser so we ignore it with the underscore
        xmltree, _ = self.load_inpxml()

        parser_info = {}
        try:
            inpxml_dict = inpxml_parser(xmltree, parser_info_out=parser_info)
        except (ValueError, FileNotFoundError) as exc:
            raise InputValidationError(f'inp.xml parser failed: {exc}') from exc
        finally:
            #Always try to provide the error/warning information
            self.parser_info = parser_info

        # set inpxml_dict attribute
        self.set_attribute('inp_dict', inpxml_dict)

    def load_inpxml(self, validate_xml_schema=True, return_included_tags=False, **kwargs):
        """
        Returns the lxml etree and the schema dictionary corresponding to the version. If validate_xml_schema=True
        the file will also be validated against the schema

        Keyword arguments are passed on to the parser
        """
        from masci_tools.io.io_fleurxml import load_inpxml
        from masci_tools.util.xml.common_functions import validate_xml

        self._validate()

        with self.open(path='inp.xml', mode='rb') as inpxmlfile:
            try:
                xmltree, schema_dict = load_inpxml(inpxmlfile, **kwargs)
            except ValueError as exc:
                # prob inp.xml file broken
                err_msg = ('The inp.xml file is probably broken, could not parse it to an xml etree.')
                raise InputValidationError(err_msg) from exc
            except FileNotFoundError as exc:
                # prob inp.xml file broken
                err_msg = ('The inp.xml file is probably broken, could not find corresponding input schema.')
                raise InputValidationError(err_msg) from exc

        xmltree, included_tags = self._include_files(xmltree)
        develop_version = self.inp_version != schema_dict['inp_version']

        if validate_xml_schema and not develop_version:
            try:
                validate_xml(xmltree,
                             schema_dict.xmlschema,
                             error_header='Input file is not validated against the schema')
            except etree.DocumentInvalid as err:
                raise InputValidationError(err) from err
        elif develop_version:
            self.logger.warning(f'You are using a Fleur input file with file version {self.inp_version}.\n'
                                'This version has no corresponding XML Schema stored in masci-tools.\n'
                                'Unexpected Errors can occur. If that is the case you can try to add the '
                                'XML Schema for this file version to masci-tools')

        if return_included_tags:
            return xmltree, schema_dict, included_tags
        else:
            return xmltree, schema_dict

    def _include_files(self, xmltree):
        """
        Tries to insert all .xml, which are not inp.xml file into the etree since they are
        not naturally available for the parser (open vs self.open)

        Creates a NamedTemporaryFile for each one with their content
        and replaces the name in the `href` attributes of the `xi:include` tags.
        Then these are expanded with `xmltree.xinclude()`
        """
        from masci_tools.util.xml.common_functions import clear_xml, eval_xpath
        import tempfile

        temp_files = []
        for file in self.files:
            #Check if the filename appears in a xi:include tag
            nodes = eval_xpath(xmltree,
                               '//xi:include[@href=$filename]',
                               namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                               filename=file,
                               list_return=True)
            if not nodes:
                continue

            #Get file content from node
            include_content = self.get_content(file)

            #Write content into temporary file
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as fo:
                fo.write(include_content)
                temp_files.append(fo.name)
                for node in nodes:
                    node.set('href', fo.name)

        #Performs the inclusions and remove comments
        cleared_tree, included_tags = clear_xml(xmltree)

        #Remove created temporary files
        for file in temp_files:
            os.remove(file)

        return cleared_tree, included_tags

    # dict with inp paramters parsed from inp.xml
    @property
    def inp_dict(self):
        """
        Returns the inp_dict (the representation of the ``inp.xml`` file) as it will
        or is stored in the database.
        """
        return self.get_attribute('inp_dict', {})

    # version of the inp.xml file
    @property
    def inp_version(self):
        """
        Returns the version string corresponding to the inp.xml file
        """
        return self.get_attribute('inp_version', None)

    def _validate(self):
        """
        A validation method. Checks if an ``inp.xml`` file is in the FleurinpData.
        """
        #from aiida.common.exceptions import ValidationError
        # check if schema file path exists.
        super()._validate()

        if 'inp.xml' in self.files:
            # has_inpxml = True # does nothing so far
            pass
        else:
            raise ValidationError('inp.xml file not in attribute "files". '
                                  'FleurinpData needs to have and inp.xml file!')

    def get_fleur_modes(self):
        """
        Analyses ``inp.xml`` file to set up a calculation mode. 'Modes' are paths a FLEUR
        calculation can take, resulting in different output.
        This files can be automatically addded to the retrieve_list of the calculation.

        Common modes are: scf, jspin2, dos, band, pot8, lda+U, eels, ...

        :return: a dictionary containing all possible modes.
        """
        from masci_tools.util.xml.xml_getters import get_fleur_modes

        xmltree, schema_dict = self.load_inpxml()

        return get_fleur_modes(xmltree, schema_dict)

    def get_nkpts(self):
        """
        This routine returns the number of kpoints used in the fleur calculation
        defined in this input

        :returns: int with the number of kPoints
        """
        from masci_tools.util.xml.xml_getters import get_nkpts

        xmltree, schema_dict = self.load_inpxml()

        return get_nkpts(xmltree, schema_dict)

    def get_structuredata_ncf(self, normalize_kind_name=True):
        """
        This routine returns an AiiDA Structure Data type produced from the ``inp.xml``
        file. not a calcfunction

        :param self: a FleurinpData instance to be parsed into a StructureData
        :returns: StructureData node, or None
        """
        from aiida.orm import StructureData
        from masci_tools.util.xml.xml_getters import get_structure_data

        xmltree, schema_dict = self.load_inpxml()

        atoms, cell, pbc = get_structure_data(xmltree,
                                              schema_dict,
                                              site_namedtuple=True,
                                              normalize_kind_name=normalize_kind_name)

        struc = StructureData(cell=cell, pbc=pbc)

        for atom in atoms:
            struc.append_atom(position=atom.position, symbols=atom.symbol, name=atom.kind)

        # TODO DATA-DATA links are not wanted, you might want to use a cf instead
        #struc.add_link_from(self, label='self.structure', link_type=LinkType.CREATE)
        # label='self.structure'
        # return {label : struc}
        return struc

    @cf
    def get_structuredata(self, normalize_kind_name=None):
        """
        This routine return an AiiDA Structure Data type produced from the ``inp.xml``
        file. If this was done before, it returns the existing structure data node.
        This is a calcfunction and therefore keeps the provenance.

        :param fleurinp: a FleurinpData instance to be parsed into a StructureData
        :returns: StructureData node
        """
        if normalize_kind_name is None:
            normalize_kind_name = Bool(True)
        return self.get_structuredata_ncf(normalize_kind_name=normalize_kind_name)

    def get_kpointsdata_ncf(self, name=None, index=None, only_used=False):
        """
        This routine returns an AiiDA :class:`~aiida.orm.KpointsData` type produced from the
        ``inp.xml`` file. This only works if the kpoints are listed in the in inpxml.
        This is NOT a calcfunction and does not keep the provenance!

        :param name: str, optional, if given only the kpoint set with the given name
                     is returned
        :param index: int, optional, if given only the kpoint set with the given index
                      is returned

        :returns: :class:`~aiida.orm.KpointsData` node
        """
        from aiida.orm import KpointsData
        from masci_tools.util.xml.xml_getters import get_kpoints_data

        # HINT, TODO:? in this routine, the 'cell' you might get in an other way
        # exp: StructureData.cell, but for this you have to make a structureData Node,
        # which might take more time for structures with lots of atoms.
        # then just parsing the cell from the inp.xml
        # as in the routine get_structureData

        xmltree, schema_dict = self.load_inpxml()

        if name is None and index is None:
            kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict, only_used=only_used)
        else:
            kpoints, weights, cell, pbc = get_kpoints_data(xmltree,
                                                           schema_dict,
                                                           name=name,
                                                           index=index,
                                                           only_used=only_used)

        if isinstance(kpoints, dict):
            kpoints_data = {}
            for (label, kpoints_set), weights_set in zip(kpoints.items(), weights.values()):
                kps = KpointsData()
                kps.set_cell(cell)
                kps.pbc = pbc
                kps.set_kpoints(kpoints_set, cartesian=False, weights=weights_set)
                #kpoints_data.add_link_from(self, label='fleurinp.kpts', link_type=LinkType.CREATE)
                pattern = re.compile(r'\W', re.UNICODE)
                kpoint_identifier = re.sub(pattern, '', label.replace('-', '_'))
                if kpoint_identifier != label.replace('-', '_'):
                    warnings.warn(
                        f'Normed the name of the kpoint set {label} to {kpoint_identifier}'
                        ' to be able to use it as a link label', UserWarning)
                kps.label = f'fleurinp.kpts.{kpoint_identifier}'
                kpoints_data[kpoint_identifier] = kps
        else:
            kpoints_data = KpointsData()
            kpoints_data.set_cell(cell)
            kpoints_data.pbc = pbc
            kpoints_data.set_kpoints(kpoints, cartesian=False, weights=weights)
            #kpoints_data.add_link_from(self, label='fleurinp.kpts', link_type=LinkType.CREATE)
            kpoints_data.label = 'fleurinp.kpts'

        return kpoints_data

    @cf
    def get_kpointsdata(self, name=None, index=None):
        """
        This routine returns an AiiDA :class:`~aiida.orm.KpointsData` type produced from the
        ``inp.xml`` file. This only works if the kpoints are listed in the in inpxml.
        This is a calcfunction and keeps the provenance!

        :returns: :class:`~aiida.orm.KpointsData` node
        """
        return self.get_kpointsdata_ncf(name=name, index=index)

    def get_parameterdata_ncf(self, inpgen_ready=True, write_ids=True):
        """
        This routine returns an AiiDA :class:`~aiida.orm.Dict` type produced from the ``inp.xml``
        file. This node can be used for inpgen as `calc_parameters`.
        This is NOT a calcfunction and does NOT keep the provenance!

        :returns: :class:`~aiida.orm.Dict` node
        """
        from aiida.orm import Dict
        from masci_tools.util.xml.xml_getters import get_parameter_data

        xmltree, schema_dict = self.load_inpxml()

        parameter_data = get_parameter_data(xmltree, schema_dict, inpgen_ready=inpgen_ready, write_ids=write_ids)

        return Dict(dict=parameter_data)

    @cf
    def get_parameterdata(self):
        """
        This routine returns an AiiDA :class:`~aiida.orm.Dict` type produced from the ``inp.xml``
        file. The returned node can be used for inpgen as `calc_parameters`.
        This is a calcfunction and keeps the provenance!

        :returns: :class:`~aiida.orm.Dict` node
        """

        return self.get_parameterdata_ncf()

    def get_tag(self, xpath):
        """
        Tries to evaluate an xpath expression for ``inp.xml`` file. If it fails it logs it.

        :param xpath: an xpath expression
        :returns: A node list retrived using given xpath
        """
        from masci_tools.util.xml.common_functions import eval_xpath

        warnings.warn(
            'The get_tag method is deprecated. Instead you can use the load_inpxml method to access '
            'the xmltree and schema of the stored inp.xml. Then the required information can be accessed '
            'via the XML functions in masci-tools or directly', DeprecationWarning)

        xmltree, _ = self.load_inpxml()
        root = xmltree.getroot()

        return eval_xpath(root, xpath)
